Large size dictionary for SKK system
Copyright (C) 1988-1995, 1997, 1999-2010

Masahiko Sato <masahiko@kuis.kyoto-u.ac.jp>
Hironobu Takahashi <takahasi@tiny.or.jp>,
Masahiro Doteguchi, Miki Inooka,
Yukiyoshi Kameyama <kameyama@kuis.kyoto-u.ac.jp>,
Akihiko Sasaki, Dai Ando, Junichi Okukawa,
Katsushi Sato and Nobuhiro Yamagishi
NAKAJIMA Mikio <minakaji@osaka.email.ne.jp>
MITA Yuusuke <clefs@mail.goo.ne.jp>
SKK Development Team <skk@ring.gr.jp>

Maintainer: SKK Development Team <skk@ring.gr.jp>
Version: $Id: SKK-JISYO.L,v 1.1089 2012/09/01 00:10:04 czkmt Exp $
Keywords: japanese
Last Modified: $Date: 2012/09/01 00:10:04 $

This dictionary is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2, or
(at your option) any later version.

This dictionary is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with Daredevil SKK, see the file COPYING.  If not, write to
the Free Software Foundation Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

